from typing import get_type_hints


def type_check(func):
    hints = get_type_hints(func)
    args_names = func.__code__.co_varnames[:func.__code__.co_argcount]

    def type_checked_func(*args, **kwargs):
        for arg, hint in hints.items():
            # We don't care about the type of the return value.
            if arg == "return":
                continue
            # Get the value of the argument.
            try:
                x = kwargs.get(arg, None)
                if not x:   
                    x = args[args_names.index(arg)]
            except IndexError:
                continue
            # Check if type matches the hint.
            if not isinstance(x, hint):
                raise TypeError(f"{arg} is {type(x)} but expected {hint}")
            
        return func(*args, **kwargs)

    return type_checked_func



from enum import Enum
from typing import Optional

from tmcp import TMCP_VERSION
from .convert import try_convert_direction
from .type_check import type_check


class ActionType(Enum):
    BALL = "BALL"
    BOOST = "BOOST"
    DEMO = "DEMO"
    READY = "READY"
    DEFEND = "DEFEND"


class TMCPMessage:
    """
    A TMCP compliant message object.
    All messages contain a `team`, `index`, and `action_type`.
    
    Based on the `action_type`, this object will have other attributes.
    For example, if the `action_type` equals `ActionType.BALL`,
    this object will have a `time` attribute.
    """

    @type_check
    def __init__(self, team: int, index: int, action_type: ActionType):
        self.team = team
        self.index = index
        self.action_type = action_type

    @classmethod
    @type_check
    def ball_action(
        cls, team: int, index: int, time: float = -1.0, direction = [0.0, 0.0, 0.0]
    ):  # -> TMCPMessage:
        self = cls(team, index, ActionType.BALL)
        self.time = time
        self.direction = try_convert_direction(direction)
        return self

    @classmethod
    @type_check
    def boost_action(cls, team: int, index: int, target: int):  # -> TMCPMessage:
        self = cls(team, index, ActionType.BOOST)
        self.target = target
        return self

    @classmethod
    @type_check
    def demo_action(
        cls, team: int, index: int, target: int, time: float = -1.0
    ):  # -> TMCPMessage:
        self = cls(team, index, ActionType.DEMO)
        self.target = target
        self.time = time
        return self

    @classmethod
    @type_check
    def ready_action(cls, team: int, index: int, time: float = -1.0):  # -> TMCPMessage:
        self = cls(team, index, ActionType.READY)
        self.time = time
        return self

    @classmethod
    @type_check
    def defend_action(cls, team: int, index: int):  # -> TMCPMessage:
        self = cls(team, index, ActionType.DEFEND)
        return self

    @classmethod
    def from_dict(cls, message: dict) -> Optional["TMCPMessage"]:
        try:
            team: int = message["team"]
            index: int = message["index"]
            assert isinstance(team, int)
            assert isinstance(index, int)

            action: dict = message["action"]
            action_type: ActionType = ActionType(action["type"])

            if action_type == ActionType.BALL:
                assert isinstance(action["time"], (float, int))
                direction = action["direction"]
                assert isinstance(direction, (list, tuple))
                assert len(direction) in (2, 3)
                assert all(isinstance(elem, (int, float)) for elem in direction)
                msg = cls.ball_action(team, index, float(action["time"]), direction)
            elif action_type == ActionType.BOOST:
                assert isinstance(action["target"], int)
                msg = cls.boost_action(team, index, action["target"])
            elif action_type == ActionType.DEMO:
                assert isinstance(action["target"], int)
                assert isinstance(action["time"], (float, int))
                msg = cls.demo_action(
                    team, index, action["target"], float(action["time"])
                )
            elif action_type == ActionType.READY:
                assert isinstance(action["time"], (float, int))
                msg = cls.ready_action(team, index, float(action["time"]))
            elif action_type == ActionType.DEFEND:
                msg = cls.defend_action(team, index)
            else:
                raise NotImplementedError
            return msg

        except (KeyError, ValueError, AssertionError):
            return None

    def to_dict(self) -> dict:
        if self.action_type == ActionType.BALL:
            action = {
                "type": "BALL",
                "time": self.time,
                "direction": self.direction
            }
        elif self.action_type == ActionType.BOOST:
            action = {
                "type": "BOOST",
                "target": self.target,
            }
        elif self.action_type == ActionType.DEMO:
            action = {
                "type": "DEMO",
                "target": self.target,
                "time": self.time,
            }
        elif self.action_type == ActionType.READY:
            action = {
                "type": "READY",
                "time": self.time,
            }
        elif self.action_type == ActionType.DEFEND:
            action = {
                "type": "DEFEND"
            }
        else:
            raise NotImplementedError

        return {
            "tmcp_version": TMCP_VERSION,
            "team": self.team,
            "index": self.index,
            "action": action,
        }

    def __repr__(self):
        return str(self.to_dict())



from time import perf_counter
from queue import Empty
from typing import List, Optional, Any

from rlbot.agents.base_agent import BaseAgent
from rlbot.matchcomms.client import MatchcommsClient

from tmcp import TMCP_VERSION
from .message import TMCPMessage


MAX_PACKETS_PER_TICK: int = 30
TIME_BETWEEN_MESSAGES: float = 0.1


class TMCPHandler:
    """The class for handling TMCP.

    Create an instance by just passing your agent in:

    ```
    def initialize_agent(self):
        self.tmcp_handler = TMCPHandler(self)
    ```

    Usage is also very straightforward:

    ```
    def get_output(self, packet: GameTickPacket) -> SimpleControllerState:
        # Receive and parse all new matchcomms messages into TMCPMessage objects.
        new_messages: List[TMCPMessage] = self.tmcp_handler.recv()
        # Handle TMCPMessages.
        for message in new_messages:
            if message.action_type == ActionType.BALL:
                print(message.time)
        
        ...

        # You can send messages like this.
        self.tmcp_handler.send_boost_action(pad_index)

        # Or you can create them and send them more directly:
        my_message = TMCPMessage.ball_action(self.team, self.index, estimated_time_of_arrival)
        self.tmcp_handler.send(my_message)
    ```
    """

    def __init__(self, agent: BaseAgent):
        self.matchcomms: MatchcommsClient = agent.matchcomms
        self.index: int = agent.index
        self.team: int = agent.team
        self.last_time: float = 0.0
        self.enabled: bool = True

    def disable(self):
        """Disable the handler. It will not send or receive any messages."""
        self.enabled = False

    def send(self, message: TMCPMessage) -> bool:
        """Send a TMCPMessage over match comms. Will not send messages if they are coming too quickly.
        Returns whether a message was sent."""

        # If disabled, pretend all messages are sent.
        # This is done so that people don't try to resend messages if the handler is disabled.
        if not self.enabled:
            return True

        current_time = perf_counter()
        if current_time - self.last_time < TIME_BETWEEN_MESSAGES:
            return False

        assert isinstance(message, TMCPMessage)
        self.matchcomms.outgoing_broadcast.put_nowait(message.to_dict())
        self.last_time: float = current_time
        return True

    def recv(self) -> List[TMCPMessage]:
        messages = []

        # Return empty message list if disabled.
        if not self.enabled:
            return messages

        # Receive messages until we reach the maximum packets per tick or the queue is empty.
        for _ in range(MAX_PACKETS_PER_TICK):
            try:
                message = self.parse(self.matchcomms.incoming_broadcast.get_nowait())
                if message is not None:
                    messages.append(message)
            except Empty:
                break
        return messages

    def parse(self, message: dict) -> Optional[TMCPMessage]:
        # Ignore messages using a different version of the protocol.
        if message.get("tmcp_version") != TMCP_VERSION:
            return None
        # Ignore messages by opposing team.
        if message.get("team") != self.team:
            return None

        return TMCPMessage.from_dict(message)

    def send_ball_action(
        self, time: Optional[float] = None, direction: Optional[Any] = None
    ) -> bool:
        """The bot is going for the ball.

        `time` - Game time that your bot will arrive at the ball.
        `direction` - Anticipated normalized direction of ball travel AFTER contact is made.
        """
        kwargs = {}
        if time is not None:
            kwargs["time"] = time
        if direction is not None:
            kwargs["direction"] = direction
        return self.send(TMCPMessage.ball_action(self.team, self.index, **kwargs))

    def send_boost_action(self, target: int) -> bool:
        """The bot is going for boost.

        `target` - Index of the boost pad the bot is going to collect.
        """
        return self.send(TMCPMessage.boost_action(self.team, self.index, target))

    def send_demo_action(self, target: int, time: Optional[float] = None) -> bool:
        """The bot is going to demolish another car.

        `target` - Index of the bot that will be demoed.
        `time` - Game time that the bot will demo the other bot.
        """
        kwargs = {}
        if time is not None:
            kwargs["time"] = time
        return self.send(
            TMCPMessage.demo_action(self.team, self.index, target, **kwargs)
        )

    def send_ready_action(self, time: Optional[float] = None) -> bool:
        """The bot is waiting for a chance to go for the ball.
        Some examples are positioning (retreating/shadowing) and recovering.

        `ready` - Game time when the bot could arrive at the ball (if it was to go for it).
        """
        kwargs = {}
        if time is not None:
            kwargs["time"] = time
        return self.send(TMCPMessage.ready_action(self.team, self.index, **kwargs))

    def send_defend_action(self) -> bool:
        """The bot is in a position to defend the goal and is not planning to move up.
        Only use DEFEND if your bot is in place to defend, not if still en-route.
        If the bot decides to leave net, signal this using either "BALL" (if going for a touch) or "READY" (if moving upfield).

        A bot should use "DEFEND" to let its teammates know it is safe to move up a bit without worrying about an open net.
        """
        return self.send(TMCPMessage.defend_action(self.team, self.index))



from typing import List, Any


def try_convert_direction(direction: Any) -> List[float]:
    """
    Attempts to convert any type into a list of 3 floats.
    Will raise a TypeError if the type is not something it can convert.
    It expects two or three coordinates (it will take z to be 0.0 if not specified).
    """
    # Convert from objects that store their coordinates in a data attribute.
    if hasattr(direction, "data"):
        direction = direction.data

    # Convert any indexable into a list, so that length can be changed if needed.
    if hasattr(direction, "__getitem__"):
        direction = list(direction)

    # Convert from objects with have x, y, z attributes.
    elif hasattr(direction, "x") and hasattr(direction, "y"):
        new_dir = [direction.x, direction.y]
        if hasattr(direction, "z"):
            new_dir.append(direction.z)
        direction = new_dir

    if isinstance(direction, list):
        if len(direction) == 2:
            direction.append(0.0)
        if len(direction) == 3:
            # Convert all to floats.
            for index, axis in enumerate(direction):
                direction[index] = float(axis)
            return direction

    raise TypeError(f"Could not convert direction, input was {direction}")



from setuptools import setup

with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

setup(
    name="tmcp",
    packages=["tmcp"],
    version="0.9.1",
    license="MIT",
    description="Helper classes for the Team Match Communication Protocol.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Viliam Vadocz",
    author_email="viliam.vadocz@gmail.com",
    url="https://github.com/ViliamVadocz/tmcp/",
    download_url="https://github.com/ViliamVadocz/tmcp/archive/v_0_9_1.tar.gz",
    keywords=["RLBot", "protocol"],
    install_requires=["rlbot"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
